package edu.georgetown.gucs.dictionary;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.IllegalDataException;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.BoundedExecutor;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.IgnoreCaseComparator;
import edu.georgetown.gucs.utility.Token;

/**
 * Creates a list of unique tokens extracted from a collection of documents that can be trimmed by removing tokens based
 * on various different attributes; used for creating fingerprints of documents that are based on words that appear in a
 * document collection
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 **/
public class Dictionary implements Serializable, Iterable<Entry<Token, DictionaryEntry>> {

  private static final long serialVersionUID = 8334786910169166049L;
  private boolean threaded;
  private boolean trained;
  private boolean training;
  private boolean trimmed;
  private int totalDocuments;
  private int version;
  private int maxThread;
  private int nextPosition;
  private double maxIDFfound;
  private double trimmed_max = 1.0;
  private double trimmed_min = 0.0;
  private String creatingProgram;
  private String creation;
  private String dictionaryFileName;
  private String dictionaryName;
  private String directory;
  private String GUID;
  private String systemID;
  private Map<Token, DictionaryEntry> dictionary = Collections
      .synchronizedSortedMap(new TreeMap<Token, DictionaryEntry>(new IgnoreCaseComparator()));
  private TokenizerList tokenizers;

  /**
   * Constructor that sets default values for this dictionary
   **/
  public Dictionary() {
    this.dictionaryName = "";
    this.version = 0;
    this.nextPosition = 0;
    this.totalDocuments = 0;
    this.trained = false;
    this.maxIDFfound = 0.0;
    setCreatingProgram();
    setSystemID();
    this.GUID = new BigInteger(128, new Random()).toString();
    this.threaded = false;
    this.maxThread = Runtime.getRuntime().availableProcessors();
    this.trimmed = false;
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "GUID=" + this.GUID);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "threaded=" + this.threaded);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "maxThread=" + this.maxThread);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "trimmed=" + this.trimmed);
  }

  /**
   * Constructor that loads this dictionary from a file
   * 
   * @param dictionaryFilename
   *          the filename containing a dictionary
   */
  public Dictionary(String dictionaryFilename) {
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "filename=" + dictionaryFilename);
    this.loadDictionaryXML(dictionaryFilename);
  }

  /**
   * Loads a previously serialized XML dictionary object into this dictionary
   * 
   * @param filename
   *          the string filename of the XML dictionary file to load
   **/
  @SuppressWarnings("unchecked")
  private synchronized void loadDictionaryXML(String filename) {
    this.dictionaryFileName = filename;

    try {
      File xml = new File(this.dictionaryFileName);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);
      Element dict = doc.getRootElement();

      this.dictionaryName = dict.getAttributeValue("name");
      this.GUID = dict.getAttributeValue("GUID");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "GUID=" + this.GUID);
      this.systemID = dict.getAttributeValue("systemID");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "systemID=" + this.systemID);
      this.version = Integer.valueOf(dict.getAttributeValue("version"));
      this.creatingProgram = dict.getAttributeValue("creatingProgram");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "creatingProgram=" + this.creatingProgram);
      this.creation = dict.getAttributeValue("creation");
      this.directory = dict.getAttributeValue("directory");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "directory=" + this.directory);
      this.maxIDFfound = Double.valueOf(dict.getAttributeValue("maxIDF"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "maxIDFfound=" + this.maxIDFfound);
      this.totalDocuments = Integer.valueOf(dict.getAttributeValue("totalDocuments"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "totalDocuments=" + this.totalDocuments);
      this.trimmed = Boolean.valueOf(dict.getAttributeValue("trimmed"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed=" + this.trimmed);
      if (this.trimmed) {
        this.trimmed_min = Double.valueOf(dict.getAttributeValue("trimmed_min"));
        Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed_min=" + this.trimmed_min);
        this.trimmed_max = Double.valueOf(dict.getAttributeValue("trimmed_max"));
        Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed_max=" + this.trimmed_max);
      }
      this.trained = Boolean.valueOf(dict.getAttributeValue("trained"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trained=" + this.trained);
      this.training = Boolean.valueOf(dict.getAttributeValue("training"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "training=" + this.training);
      this.nextPosition = Integer.valueOf(dict.getAttributeValue("nextPosition"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "nextPosition=" + this.nextPosition);

      Vector<String> tokenizerNames = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }
      setTokenizers(tokenizerNames);
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "tokenizers=" + this.tokenizers);

      synchronized (this.dictionary) {
        if (!this.dictionary.isEmpty()) {
          this.dictionary.clear();
        }

        for (Element tokenElement : (List<Element>) dict.getChildren("token")) {
          Token token = new Token(tokenElement.getAttributeValue("name"), Integer.valueOf(tokenElement
              .getAttributeValue("start_byte")), Integer.valueOf(tokenElement.getAttributeValue("end_byte")));
          this.dictionary.put(
              token,
              new DictionaryEntry(token, Integer.valueOf(tokenElement.getAttributeValue("position")), Integer
                  .valueOf(tokenElement.getAttributeValue("frequency")), Integer.valueOf(tokenElement
                  .getAttributeValue("documentCount")), Double.valueOf(tokenElement.getAttributeValue("idf"))));
        }
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Reads an XML file that specifies which tokenizers to use for this dictionary
   * 
   * @param config_file
   *          the string name of the XML configuration file containing this dictionary's tokenizers
   **/
  @SuppressWarnings("unchecked")
  private void addTokenizers(String config_file) {
    Global.monitor.writeToBuffer("Dictionary", "addTokenizers", "config_file=" + config_file);
    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);

      Element dict = doc.getRootElement();
      Vector<String> tokenizerNames = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }
      setTokenizers(tokenizerNames);
      Global.monitor.writeToBuffer("Dictionary", "addTokenizers", "tokenizers=" + this.tokenizers);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Adds a document to this dictionary. Requires this dictionary to be in training mode.
   * 
   * @param docTokens
   *          all tokens in the document
   **/
  private void addDocument(List<Token> docTokens) {
    if (this.training) {
      // iterate over all terms and insert them into a temp hash map
      Map<Token, DictionaryEntry> file_input = Collections.synchronizedSortedMap(new TreeMap<Token, DictionaryEntry>(
          new IgnoreCaseComparator()));
      synchronized (file_input) {
        for (Token token : docTokens) {
          if (file_input.containsKey(token)) {
            file_input.get(token).incrementFrequencyCount();
          } else {
            DictionaryEntry toInsert = new DictionaryEntry(token);
            file_input.put(token, toInsert);
          }
        }
        mergeDictionary(file_input);
      }
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("Not in training mode"));
    }
  }

  /**
   * Creates a deep copy of the given dictionary object and stores it in this dictionary
   * 
   * @param anotherDictionary
   *          the dictionary object to copy
   */
  public void clone(Dictionary anotherDictionary) {
    this.dictionary.clear();
    this.dictionary.putAll(anotherDictionary.dictionary);
    this.directory = anotherDictionary.directory;
    Global.monitor.writeToBuffer("Dictionary", "clone", "directory=" + this.directory);
    this.nextPosition = anotherDictionary.nextPosition;
    Global.monitor.writeToBuffer("Dictionary", "clone", "nextPosition=" + this.nextPosition);
    this.trained = anotherDictionary.trained;
    Global.monitor.writeToBuffer("Dictionary", "clone", "trained=" + this.trained);
    this.training = anotherDictionary.training;
    Global.monitor.writeToBuffer("Dictionary", "clone", "training=" + this.training);
    this.version = anotherDictionary.version;
    Global.monitor.writeToBuffer("Dictionary", "clone", "version=" + this.version);
    this.trimmed = anotherDictionary.trimmed;
    Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed=" + this.trimmed);
    if (this.trimmed) {
      this.trimmed_min = anotherDictionary.trimmed_min;
      Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed_min=" + this.trimmed_min);
      this.trimmed_max = anotherDictionary.trimmed_max;
      Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed_max=" + this.trimmed_max);
    }
    this.maxIDFfound = anotherDictionary.maxIDFfound;
    Global.monitor.writeToBuffer("Dictionary", "clone", "maxIDFfound=" + this.maxIDFfound);
    this.totalDocuments = anotherDictionary.totalDocuments;
    Global.monitor.writeToBuffer("Dictionary", "clone", "totalDocuments=" + this.totalDocuments);
    this.tokenizers = anotherDictionary.tokenizers;
    Global.monitor.writeToBuffer("Dictionary", "clone", "tokenizers=" + this.tokenizers);
  }

  /**
   * Determines if this dictionary contains a particular token
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>boolean</code> value of whether the token is present in this dictionary
   **/
  public boolean containsToken(Token token) {
    return this.dictionary.containsKey(token);
  }

  /**
   * Creates this dictionary using the entire fileLister provided
   * 
   * @param fileLister
   *          the fileLister containing the files to use for this dictionary
   */
  private void createFullDictionary(FileLister fileLister) {
    synchronized (this.dictionary) {
      this.dictionary.clear();
    }
    this.directory = fileLister.getDirectory().getAbsolutePath();

    startTraining();
    Global.monitor.printMessage("Dictionary", "createFullDictionary: threaded=" + this.threaded);
    Global.monitor.writeToBuffer("Dictionary", "createFullDictionary", "threaded=" + this.threaded);

    if (this.threaded) {
      // Use multiple threads to speed execution of creating this dictionary
      ExecutorService executorService = Executors.newFixedThreadPool(this.maxThread);
      BoundedExecutor boundExecutor = new BoundedExecutor(executorService, this.maxThread);

      while (!fileLister.loadComplete()) {
        DictionaryWorker currentFileProcessor = new DictionaryWorker(fileLister.takeFile(), this.tokenizers, this);
        boundExecutor.submitTask(currentFileProcessor);
      }

      executorService.shutdown();
      while (!executorService.isTerminated()) {
        try {
          // Wait a second and see if we are done
          Thread.sleep(1000);
        } catch (InterruptedException e) {
          ExceptionHandler.handleExceptionExit(e);
        }
      }

    } else {
      Global.monitor.writeToBuffer("Dictionary", "createFromFileLister", "adding documents");
      while (!fileLister.loadComplete()) {

        for (Entry<String, List<Token>> splitterEntry : this.tokenizers.tokenizeFile(
            fileLister.takeFile().getAbsolutePath()).entrySet()) {
          if (splitterEntry.getValue() != null) {
            this.addDocument(splitterEntry.getValue());
          }
        }
      }
      Global.monitor.writeToBuffer("Dictionary", "createFromFileLister", "documents added");
    }

    endTraining();
  }

  /**
   * Creates this dictionary using the optional count or percent in the given fileLister; if no count or percent is
   * specified then uses entire fileLister to create dictionary
   * 
   * @param fileLister
   *          the fileLister containing the files to use for this dictionary
   */
  public void createPartialDictionary(FileLister fileLister) {
    synchronized (this.dictionary) {
      this.dictionary.clear();
    }
    this.directory = fileLister.getDirectory().getAbsolutePath();

    startTraining();
    Global.monitor.printMessage("Dictionary", "createPartialDictionary: threaded=" + this.threaded);
    Global.monitor.writeToBuffer("Dictionary", "createPartialDictionary", "threaded=" + this.threaded);
    Iterator<File> dictionary_iterator = fileLister.dictionary_iterator();

    if (this.threaded) {
      // Use multiple threads to speed execution of creating this dictionary
      ExecutorService executorService = Executors.newFixedThreadPool(this.maxThread);
      BoundedExecutor boundExecutor = new BoundedExecutor(executorService, this.maxThread);

      while (dictionary_iterator.hasNext()) {
        DictionaryWorker currentFileProcessor = new DictionaryWorker(dictionary_iterator.next(), this.tokenizers, this);
        boundExecutor.submitTask(currentFileProcessor);
      }

      executorService.shutdown();
      while (!executorService.isTerminated()) {
        try {
          // Wait a second and see if we are done
          Thread.sleep(1000);
        } catch (InterruptedException e) {
          ExceptionHandler.handleExceptionExit(e);
        }
      }
      Global.monitor.printMessage("Dictionary", "createPartialDictionary: after shutdown");

    } else {
      Global.monitor.printMessage("Dictionary", "createPartialDictionary: adding documents");
      Global.monitor.writeToBuffer("Dictionary", "createPartialDictionary", "adding documents");
      while (dictionary_iterator.hasNext()) {
        for (Entry<String, List<Token>> splitterEntry : this.tokenizers.tokenizeFile(
            dictionary_iterator.next().getAbsolutePath()).entrySet()) {
          if (splitterEntry.getValue() != null) {
            this.addDocument(splitterEntry.getValue());
          }
        }
      }
      Global.monitor.writeToBuffer("Dictionary", "createPartialDictionary", "documents added");
    }
    Global.monitor.printMessage("Dictionary", "createPartialDictionary ended");
    endTraining();
  }

  /**
   * Allows for operations on the tokens in this dictionary
   * 
   * @return the iterator over the tokens and their dictionaryEntry in this dictionary
   **/
  @Override
  public Iterator<Entry<Token, DictionaryEntry>> iterator() {
    synchronized (this.dictionary) {
      return this.dictionary.entrySet().iterator();
    }
  }

  /**
   * Starts training, which permits tokens to be added to this dictionary
   **/
  private void startTraining() {
    Global.monitor.writeToBuffer("Dictionary", "startTraining", "");
    if (this.trimmed) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Training a trimmed dictionary is a bad idea"));
    }
    this.training = true;
  }

  /**
   * Ends training, which prohibits adding new tokens to this dictionary and allows this dictionary to be written to a
   * file. Increments the version number and calculates the largest IDF in this dictionary.
   **/
  private void endTraining() {
    if (this.training) {
      this.version++;
      this.trained = true;
      this.training = false;
      computeIDFs();
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("Not in training mode"));
    }
    Global.monitor.writeToBuffer("Dictionary", "endTraining", "");
  }

  /**
   * Computes the Inverse Document Frequencies (IDFs) for each token in the dictionary and determines the largest IDF
   **/
  private void computeIDFs() {
    for (Entry<Token, DictionaryEntry> token : this) {
      double idf = token.getValue().computeIDF(this.totalDocuments);
      this.maxIDFfound = (idf > this.maxIDFfound) ? idf : this.maxIDFfound;
    }
    Global.monitor.writeToBuffer("Dictionary", "findMaxIDF", "maxIDFfound=" + this.maxIDFfound);
  }

  /**
   * Provides the program that created this dictionary
   * 
   * @return the string name of the program that created this dictionary
   **/
  public String getCreatingProgram() {
    return this.creatingProgram;
  }

  /**
   * Provides the maximum thread count to use for this dictionary
   * 
   * @return the maximum thread count to use for this dictionary
   **/
  public int getMaxThread() {
    return this.maxThread;
  }

  /**
   * Provides the original creation date of this dictionary
   * 
   * @return the string date when this dictionary was created
   **/
  public String getCreation() {
    return this.creation;
  }

  /**
   * Provides the filename of this dictionary, if this dictionary was loaded from or saved to a file
   * 
   * @return the string dictionary filename
   **/
  public String getDictionaryFilename() {
    return this.dictionaryFileName;
  }

  /**
   * Provides the name of this dictionary
   * 
   * @return the string name of this dictionary
   **/
  public String getDictionaryName() {
    return this.dictionaryName;
  }

  /**
   * Provides the unique identifier for this dictionary
   * 
   * @return the string GUID for dictionary
   **/
  public String getGUID() {
    return this.GUID;
  }

  /**
   * Provides the largest IDF value in this dictionary
   * 
   * @return the <code>double</code> value of the largest IDF in this dictionary
   */
  public double getMaxIDFFound() {
    return this.maxIDFfound;
  }

  /**
   * Provides the largest IDF value in this dictionary
   * 
   * @return the <code>double</code> value of the largest IDF in this dictionary
   */
  public double getMaxIDF() {
    return this.trimmed_max;
  }

  /**
   * Provides the largest IDF value in this dictionary
   * 
   * @return the <code>double</code> value of the largest IDF in this dictionary
   */
  public double getMinIDF() {
    return this.trimmed_min;
  }

  /**
   * Provides the normalized inverse document frequency (IDF) of the given token in this dictionary. The normalized IDF
   * is the token's IDF divided by the highest IDF in this dictionary.
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>double</code> value of the token's normalized inverse document frequency (IDF); returns 0.0 if
   *         the token is not present in this dictionary
   **/
  public double getNormalizedIDF(Token token) {
    if (!this.training) {
      synchronized (this.dictionary) {
        if (this.dictionary.containsKey(token)) {
          return getNormalizedIDF(this.dictionary.get(token));
        }
      }
      ExceptionHandler.handleExceptionThrow(new Exception("Token not in dictionary"));
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary in training mode"));
    }
    return 0.0;
  }

  /**
   * Provides the normalized inverse document frequency (IDF) of the given dictionaryEntry. The normalized IDF is the
   * token's IDF divided by the highest IDF in this dictionary.
   * 
   * @param entry
   *          the dictionaryEntry to calculate the normalized IDF
   * 
   * @return the <code>double</code> value of the dictionaryEntry's normalized inverse document frequency (IDF)
   */
  public double getNormalizedIDF(DictionaryEntry entry) {
    return entry.getIDF() / this.maxIDFfound;
  }

  /**
   * Provides the position of the given token in this dictionary
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the position of the token; returns -1 if the token is not present in this dictionary
   **/
  public int getPosition(Token token) {
    if (!this.training) {
      synchronized (this.dictionary) {
        if (this.dictionary.containsKey(token)) {
          return this.dictionary.get(token).getPosition();
        }
      }
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary in training mode"));
    }
    return -1;
  }

  /**
   * Provides the number of tokens in this dictionary
   * 
   * @return the number of unique tokens in this dictionary
   **/
  public int size() {
    synchronized (this.dictionary) {
      return this.dictionary.size();
    }
  }

  /**
   * Provides the directory used for this dictionary
   * 
   * @return the string directory for which this dictionary was created
   **/
  public String getSource() {
    return this.directory;
  }

  /**
   * Provides the identifier for the system that this dictionary was created on
   * 
   * @return the string system ID for this dictionary
   **/
  public String getSystemID() {
    return this.systemID;
  }

  /**
   * Provides the tokenizerList object used to create this dictionary
   * 
   * @return the tokenizerList object used for this dictionary
   **/
  public TokenizerList getTokenizers() {
    return this.tokenizers;
  }

  /**
   * Provides a vector of the tokens in this dictionary
   * 
   * @return the list of tokens in this dictionary
   **/
  public synchronized List<Token> getTokens() {
    synchronized (this.dictionary) {
      return Collections.synchronizedList(new Vector<Token>(this.dictionary.keySet()));
    }
  }

  /**
   * Provides the number of documents processed for this dictionary
   * 
   * @return the total number of documents used to create this dictionary
   **/
  public int getTotalDocuments() {
    return this.totalDocuments;
  }

  /**
   * Provides the version number of this dictionary. The version is incremented when training on this dictionary ends.
   * 
   * @return the version number of this dictionary
   **/
  public int getVersion() {
    return this.version;
  }

  /**
   * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs
   * dictionary to a file
   * 
   * @param config_file
   *          the string name of the XML configuration file containing dictionary tokenizers
   * @param path
   *          the string path to the directory used to create this dictionary
   * @param output_file
   *          the string name/path to write this dictionary to
   */
  public void makeDictionary(String config_file, String path, String output_file) {
    if (output_file == null || config_file == null || path == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null parameter(s) given."));
    }

    addTokenizers(config_file);

    if (this.tokenizers.getNames() == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null tokenizerList"));
    }
    if (this.tokenizers.getNames().size() < 1) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Empty tokenizerList"));
    }
    if (this.dictionary == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null dictionary"));
    }

    threadingOn();

    FileLister files = new FileLister(path, this.maxThread);
    createFullDictionary(files);

    saveDictionaryXML(output_file);
  }

  /**
   * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs a
   * trimmed dictionary to a file
   * 
   * @param tokenizersNames
   *          the list of names of the dictionary tokenizers
   * @param path
   *          the string path to the directory used to create this dictionary
   * @param min_idf
   *          the <code>double</code> minimum normalized inverse document frequency (IDF) to trim dictionary
   * @param max_idf
   *          the <code>double</code> maximum normalized inverse document frequency (IDF) to trim dictionary
   * @param output_file
   *          the string name/path to write this dictionary to
   */
  public void makeDictionary(List<String> tokenizersNames, String path, double min_idf, double max_idf,
      String output_file) {
    if (output_file == null || tokenizersNames == null || path == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null parameter(s) given."));
    }

    setTokenizers(tokenizersNames);

    if (this.tokenizers.getNames() == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null tokenizerList"));
    }
    if (this.tokenizers.getNames().size() < 1) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Empty tokenizerList"));
    }
    if (this.dictionary == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null dictionary"));
    }

    // synchronized (this.dictionary) {
    // this.dictionary = Collections.synchronizedSortedMap(new TreeMap<Token, DictionaryEntry>(
    // new IgnoreCaseComparator()));
    // }
    threadingOn();

    FileLister files = new FileLister(path, this.maxThread);
    createFullDictionary(files);

    if (min_idf > 0 || max_idf < 1.0) {
      trimByIDF(min_idf, max_idf);
    }
    saveDictionaryXML(output_file);
  }

  /**
   * Merges the given Map of tokens to dictionaryEntry with this dictionary
   * 
   * @param documentTokens
   *          the map containing the dictionary information to merge with this dictionary
   */
  protected void mergeDictionary(Map<Token, DictionaryEntry> documentTokens) {
    synchronized (this.dictionary) {
      for (Entry<Token, DictionaryEntry> documentEntry : documentTokens.entrySet()) {
        if (this.dictionary.containsKey(documentEntry.getKey())) {
          // If so, increment the document count and frequency count
          this.dictionary.get(documentEntry.getKey()).incrementFrequencyCount(
              documentEntry.getValue().getFrequencyCount());
          this.dictionary.get(documentEntry.getKey()).incrementDocumentCount();
        } else {
          // Otherwise, insert the DictionaryEntry into this dictionary
          documentEntry.getValue().setPosition(this.nextPosition);
          this.nextPosition++;
          this.dictionary.put(documentEntry.getKey(), documentEntry.getValue());
        }
      }
      this.totalDocuments++;
    }
  }

  /**
   * Writes this dictionary as a serialized XML object. Requires dictionary to be out of training mode.
   * 
   * @param filename
   *          the string XML filename to write this dictionary to
   **/
  void saveDictionaryXML(String filename) {
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "threaded=" + this.threaded);
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "maxThread=" + this.maxThread);
    if (!this.training) {
      this.dictionaryFileName = (new File(filename)).getAbsolutePath();
      this.dictionaryName = filename.substring(filename.lastIndexOf('/') + 1);
      this.creation = String.valueOf(new Date());
      DecimalFormat rounding = new DecimalFormat("#0.####");

      Element dict = new Element("dictionary");
      dict.setAttribute("name", this.dictionaryFileName);
      dict.setAttribute("GUID", this.GUID);
      dict.setAttribute("systemID", this.systemID);
      dict.setAttribute("version", String.valueOf(this.version));
      if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
        dict.setAttribute("creatingProgram", this.creatingProgram);
      if (getCreation() != null && getCreation().length() > 0)
        dict.setAttribute("creation", this.creation);
      if (getSource() != null && getSource().length() > 0)
        dict.setAttribute("directory", this.directory);
      dict.setAttribute("maxIDF", rounding.format((this.maxIDFfound)));
      dict.setAttribute("totalDocuments", String.valueOf(this.totalDocuments));
      dict.setAttribute("trimmed", String.valueOf(this.trimmed));
      if (this.trimmed) {
        dict.setAttribute("trimmed_min", String.valueOf(this.trimmed_min));
        dict.setAttribute("trimmed_max", String.valueOf(this.trimmed_max));
      }
      dict.setAttribute("trained", String.valueOf(this.trained));
      dict.setAttribute("training", String.valueOf(this.training));
      dict.setAttribute("nextPosition", String.valueOf(this.nextPosition));

      for (String tokenizer : this.tokenizers.getNames()) {
        String[] split = tokenizer.split("\\s");
        Element tokenizerElement = new Element("tokenizer");
        tokenizerElement.setAttribute("name", split[0]);
        StringBuffer parameters = new StringBuffer();
        if (split.length > 1) {
          parameters.append(split[1]);
          tokenizerElement.setAttribute("parameter", split[1]);
        }
        dict.addContent(tokenizerElement);
      }

      for (Entry<Token, DictionaryEntry> entry : this) {
        try {
          Element tokenElement = new Element("token");
          tokenElement.setAttribute("name", entry.getKey().name());
          tokenElement.setAttribute("start_byte", String.valueOf(entry.getKey().start_byte()));
          tokenElement.setAttribute("end_byte", String.valueOf(entry.getKey().end_byte()));
          tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
          tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
          tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
          tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
          tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
          dict.addContent(tokenElement);
        } catch (IllegalDataException exception) {
          Global.monitor.printErrorMessage("Dictionary",
              "org.jdom.IllegalDataException prohibits token from being added to dictionary: " + entry.getKey());
          Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML",
              "org.jdom.IllegalDataException prohibits token from being added to dictionary: " + entry.getKey());
        }
      }

      try {
        Document doc = new Document(dict);
        FileOutputStream dictionaryOut = new FileOutputStream(this.dictionaryFileName);
        Format form = Format.getPrettyFormat();
        form.setEncoding("UTF-8");
        XMLOutputter out = new XMLOutputter(form);
        out.output(doc, dictionaryOut);
        System.out.println("Writing dictionary to " + this.dictionaryFileName);
        Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "Writing dictionary to "
            + this.dictionaryFileName);
        dictionaryOut.close();
      } catch (FileNotFoundException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("No saves in training mode."));
    }
  }

  /**
   * Determines the name of the program that created this dictionary
   **/
  private void setCreatingProgram() {
    Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "start");
    this.creatingProgram = "";
    Runtime runtime = Runtime.getRuntime();
    String cmds[] = { "jps", "-m" };
    Process proc;
    try {
      proc = runtime.exec(cmds);
      BufferedReader bufferedreader = new BufferedReader(new InputStreamReader(proc.getInputStream(), "UTF-8"));

      String line;
      while ((line = bufferedreader.readLine()) != null) {
        for (String output : line.split("\\s")) {
          if (output.contains(".jar")) {
            this.creatingProgram = output;
            Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "creatingProgram=" + this.creatingProgram);
          }
        }
      }
      bufferedreader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Sets the name of the program that created this dictionary
   * 
   * @param creatingProgram
   *          the string program that created this dictionary
   */
  public void setCreatingProgram(String creatingProgram) {
    this.creatingProgram = creatingProgram;
    Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "creatingProgram=" + creatingProgram);
  }

  /**
   * Sets the system ID of this dictionary
   */
  private void setSystemID() {
    try {
      this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    Global.monitor.writeToBuffer("Dictionary", "setSystemID", "systemID=" + this.systemID);
  }

  /**
   * Loads the tokenizers to use for this dictionary from a TokenizerList object. Only works if no tokenizers have
   * already been set.
   * 
   * @param tokenizerList
   *          the TokenizerList object containing the tokenizers to use for this dictionary
   */
  public void setTokenizers(TokenizerList tokenizerList) {
    if (this.tokenizers == null) {
      this.tokenizers = tokenizerList;
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Cannot reset tokenizer list in a dictionary"));
    }
    Global.monitor.writeToBuffer("Dictionary", "setTokenizers", "tokenizers=" + this.tokenizers);
  }

  /**
   * Sets the tokenizers used by this dictionary; Only works if no tokenizers have already been set or if the given list
   * contains the same tokenizers as those that have already been set
   * 
   * @param tokenizerVec
   *          the list of tokenizer names to use to create this dictionary
   **/
  public void setTokenizers(List<String> tokenizerVec) {
    Global.monitor.writeToBuffer("Dictionary", "setTokenizers", "tokenizers=" + tokenizerVec);
    if (this.tokenizers == null) {
      this.tokenizers = new TokenizerList(tokenizerVec);
    } else if (this.tokenizers.getNames().size() != tokenizerVec.size()
        || !tokenizerVec.containsAll(this.tokenizers.getNames())) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Cannot reset tokenizers in a dictionary"));
    }
    Global.monitor.writeToBuffer("Dictionary", "setTokenizers", "tokenizers=" + this.tokenizers);
  }

  /**
   * Prints this dictionary statistics, including:
   * <ul>
   * <li>Name</li>
   * <li>Number of documents</li>
   * <li>Number of tokens</li>
   * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
   * <li>Minimum and maximum IDF</li>
   * <li>List of tokenizers</li>
   * </ul>
   **/
  void showStatistics() {
    System.out.println("Name: " + this.dictionaryFileName);
    System.out.println("# Docs: " + this.totalDocuments);
    synchronized (this.dictionary) {
      System.out.println("# Tokens: " + this.dictionary.size());
    }
    if (this.trimmed) {
      System.out
          .println("Trimmed: " + this.trimmed + " (min: " + this.trimmed_min + ", max: " + this.trimmed_max + ")");
    } else {
      System.out.println("Trimmed: " + this.trimmed);
    }
    System.out.println("Max IDF: " + this.maxIDFfound);
    System.out.println("Tokenizers: " + this.tokenizers);
  }

  /**
   * Provides this dictionary statistics, including:
   * <ul>
   * <li>Name</li>
   * <li>Number of documents</li>
   * <li>Number of tokens</li>
   * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
   * <li>Minimum and maximum IDF</li>
   * <li>List of tokenizers</li>
   * </ul>
   * 
   * @return the string containing this dictionary's statistics
   **/
  public String getStatistics() {
    StringBuffer statisticsBuffer = new StringBuffer();
    statisticsBuffer.append(String.format("%s%n", "Name: " + this.dictionaryFileName));
    statisticsBuffer.append(String.format("%s%n", "# Docs: " + this.totalDocuments));
    synchronized (this.dictionary) {
      statisticsBuffer.append(String.format("%s%n", "# Tokens: " + this.dictionary.size()));
    }
    if (this.trimmed) {
      statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed + " (min: " + this.trimmed_min
          + ", max: " + this.trimmed_max + ")"));
    } else {
      statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed));
    }
    statisticsBuffer.append(String.format("%s%n", "Max IDF: " + this.maxIDFfound));
    statisticsBuffer.append(String.format("%s%n", "Tokenizers: " + this.tokenizers));
    return statisticsBuffer.toString();
  }

  /**
   * Prints each token in this dictionary with it's frequency, IDF, and normalized IDF
   **/
  void showTokens() {
    System.out.format("%-25s %-10s %-10s %-10s%n", "Token", "Freq", "IDF", "Norm. IDF");
    System.out.println("----------------------------------------------------------");
    for (Entry<Token, DictionaryEntry> entry : this) {
      System.out.format("%-25s %-10s %-10.4f %-10.4f%n", entry.getKey(), entry.getValue().getFrequencyCount(), entry
          .getValue().getIDF(), getNormalizedIDF(entry.getValue()));
    }
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   **/
  private void threadingOn() {
    this.threaded = true;
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "threaded=" + this.threaded);
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   * 
   * @param numThreads
   *          the number of threads to use to create this dictionary
   **/
  public void threadingOn(int numThreads) {
    this.maxThread = numThreads;
    this.threaded = true;
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "threaded=" + this.threaded);
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "maxThread=" + this.maxThread);
  }

  /**
   * Trims this dictionary by removing any token that is outside a range of normalized IDFs. Once trimmed, no new tokens
   * can be added to this dictionary.
   * 
   * @param min
   *          the <code>double</code> minimum normalized IDF to keep; any token with a lower IDF will be discarded.
   * @param max
   *          the <code>double</code> maximum normalized IDF to keep; any token with a higher IDF will be discarded.
   **/
  public void trimByIDF(double min, double max) {
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "start");

    if (min >= max) {
      ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Invalid IDF range:  min_idf=" + min
          + " max_idf=" + max));
    }

    // Trims this dictionary to contain only tokens within the specified range of normalized IDF values
    if (this.trimmed) {
      Global.monitor.printErrorMessage("Dictionary", "Trimming an already trimmed dictionary might be a bad idea");
      Global.monitor.writeToBuffer("Dictionary", "trimByIDF",
          "Trimming an already trimmed dictionary might be a bad idea");
    }

    this.trimmed = true;
    this.trimmed_min = min;
    this.trimmed_max = max;
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed=" + this.trimmed);
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed_min=" + this.trimmed_min);
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed_max=" + this.trimmed_max);
    Global.monitor.printMessage("Dictionary", "trimByIDF: trimmed=" + this.trimmed + "  min=" + min + "  max=" + max);

    int removedTokens = 0;
    synchronized (this.dictionary) {
      Iterator<Entry<Token, DictionaryEntry>> iterator = this.iterator();
      while (iterator.hasNext()) {
        Entry<Token, DictionaryEntry> entry = iterator.next();
        if (getNormalizedIDF(entry.getValue()) < min || getNormalizedIDF(entry.getValue()) > max) {
          iterator.remove();
          removedTokens++;
        } else {
          entry.getValue().setPosition(entry.getValue().getPosition() - removedTokens);
          this.dictionary.put(entry.getKey(), entry.getValue());
        }
      }
    }
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "end");
  }

  /**
   * Creates a dictionary from the specified directory using the provided tokenizers
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the tokenizer configuration file<br>
   *          <code>args[1]</code> the directory to create the dictionary from<br>
   *          <code>args[2]</code> the filename to write the dictionary to
   */
  public static void main(String args[]) {
    String output_file = null;
    String directory = null;
    String config_file = null;

    try {
      config_file = args[0];
      directory = args[1];
      output_file = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CreateDictionary <tokenizers config file> <directory> <output dictionary filename> ");
      System.err.println("Recursively creates a dictionary from the specified directory using the provided tokenizers");
      System.exit(-1);
    }
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "config=" + config_file);
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "directory=" + directory);
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "output=" + output_file);

    try {
      Dictionary dict = new Dictionary();
      dict.makeDictionary(config_file, directory, output_file);
      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
