package edu.georgetown.gucs.experiment;

import java.io.File;
import java.util.List;
import java.util.Random;

import org.apache.xerces.impl.dv.util.Base64;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.utility.Token;

/**
 * Provides the trial parameters that are needed for results comparisons
 * 
 * @author Lindsay Neubauer
 */
class TrialParameters {

  private final int experimentID;
  private final boolean insert;
  private final String dataset;
  private final double sampleParameter;
  private final double dictionaryParameter;
  private final List<File> sampleFiles;
  private final DBInterface database;
  private final Fingerprinter fingerprinter;
  private final FingerprintMatcher matcher;
  private final Dictionary dictionary;
  private final String mangler;
  private final String splitter;

  private Token file_token;
  private byte[] fingerprint;
  private long randomSeed;
  private Random random;

  /**
   * Constructor that sets default values for a trial
   * 
   * @param experimentID
   *          the <code>int</code> value of the experiment id this trial is a part of; set to -1 if insert = false
   * @param insert
   *          the <code>boolean</code> value of whether to insert the results from this trial into the database
   * @param database
   *          the dBInterface database to insert trial results; null if insert = false
   * @param dataset
   *          the string name of the database to insert trial results
   * @param sampleFiles
   *          the list of sample files to use in this trial
   * @param dictionaryParameter
   *          the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param sampleParameter
   *          the <code>double</code> containing the count or percent to use for the file sample size in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param dictionary
   *          the dictionary used for this trial
   * @param fingerprinter
   *          the fingerprinter to use in this trial
   * @param matcher
   *          the matcher to use for fingerprint comparisons in this trial
   * @param mangler
   *          the string containing the mangler(s)
   * @param splitter
   *          the string containing the splitter
   */
  protected TrialParameters(int experimentID, boolean insert, DBInterface database, String dataset,
      List<File> sampleFiles, double dictionaryParameter, double sampleParameter, Dictionary dictionary,
      Fingerprinter fingerprinter, FingerprintMatcher matcher, String mangler, String splitter) {

    this.experimentID = experimentID;
    this.insert = insert;
    this.database = database;
    this.dataset = dataset;
    this.sampleFiles = sampleFiles;
    this.dictionaryParameter = dictionaryParameter;
    this.sampleParameter = sampleParameter;
    this.dictionary = dictionary;
    this.fingerprinter = fingerprinter;
    this.matcher = matcher;
    this.mangler = mangler;
    this.splitter = splitter;
  }

  protected TrialParameters(TrialParameters trialParameters) {

    this.experimentID = trialParameters.getExperimentID();
    this.insert = trialParameters.isInsert();
    this.database = trialParameters.getDatabase();
    this.dataset = trialParameters.getDataset();
    this.sampleFiles = trialParameters.getSampleFiles();
    this.dictionaryParameter = trialParameters.getDictionaryParameter();
    this.sampleParameter = trialParameters.getSampleParameter();
    this.dictionary = trialParameters.getDictionary();
    this.fingerprinter = trialParameters.getFingerprinter();
    this.matcher = trialParameters.getMatcher();
    this.mangler = trialParameters.getMangler();
    this.splitter = trialParameters.getSplitter();
  }

  /**
   * Sets the non-final information for this trial parameters
   * 
   * @param file_token
   *          the token containing file name and its byte information
   * @param fingerprint
   *          the byte array containing the fingerprint
   */
  protected void set(Token file_token, byte[] fingerprint) {
    this.file_token = file_token;
    this.fingerprint = fingerprint;
  }

  /**
   * Sets the random seed for the random number generator
   * 
   * @param randomSeed
   *          the <code>long</code> the initial seed to use for the random number generator in this trial; can be used
   *          to enable repeatability across trials
   */
  public void setRandomSeed(long randomSeed) {
    this.randomSeed = randomSeed;
    this.random = new Random(randomSeed);
  }

  /**
   * Provides the token containing file information
   * 
   * @return the token file
   */
  public Token getFileToken() {
    return this.file_token;
  }

  /**
   * Provides the fingerprint
   * 
   * @return the byte array containing the fingerprint
   */
  public byte[] getFingerprint() {
    return this.fingerprint;
  }

  /**
   * Provides the experiment ID for the trial
   * 
   * @return the <code>int</code> value of the experiment id this trial is a part of; set to -1 if insert = false
   */
  public int getExperimentID() {
    return this.experimentID;
  }

  /**
   * Provides the boolean value of whether results are inserted into the database
   * 
   * @return the <code>boolean</code> value of whether to insert the results from this trial into the database
   */
  public boolean isInsert() {
    return this.insert;
  }

  /**
   * Provides the name of the dataset
   * 
   * @return the string name of the database to insert trial results
   */
  public String getDataset() {
    return this.dataset;
  }

  /**
   * Provides the list of files used for the sample
   * 
   * @return the list of sample files to use in this trial
   */
  public List<File> getSampleFiles() {
    return this.sampleFiles;
  }

  /**
   * Provides the count or percent to use for the trial's sample
   * 
   * @return the <code>double</code> containing the count or percent to use for the file sample size in this trial;
   *         numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   */
  public double getSampleParameter() {
    return this.sampleParameter;
  }

  /**
   * Provides the count or percent to use for the trial's dictionary
   * 
   * @return the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
   *         numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   */
  public double getDictionaryParameter() {
    return this.dictionaryParameter;
  }

  /**
   * Provides the name of the splitter
   * 
   * @return the string name of the splitter to use in this trial
   */
  public String getSplitter() {
    return this.splitter;
  }

  /**
   * Provides the name of the manglers
   * 
   * @return the string name of the manglers to use in this trial
   */
  public String getMangler() {
    return this.mangler;
  }

  /**
   * Provides the initial seed to use for the random number generator in this trial
   * 
   * @return the initial seed for the random number generator in this trial; can be used to enable repeatability across
   *         trials
   */
  public long getRandomSeed() {
    return this.randomSeed;
  }

  /**
   * Provides the random number generator
   * 
   * @return the random number generator in this trial; can be used to enable repeatability across trials
   */
  public Random getRandom() {
    return this.random;
  }

  /**
   * Provides database containing trial results
   * 
   * @return the dBInterface database to insert trial results; null if insert = false
   */
  public DBInterface getDatabase() {
    return this.database;
  }

  /**
   * Provides the fingerprinter
   * 
   * @return the fingerprinter to use in this trial
   */
  public Fingerprinter getFingerprinter() {
    return this.fingerprinter;
  }

  /**
   * Provides the matcher
   * 
   * @return the matcher to use for fingerprint comparisons in this trial
   */
  public FingerprintMatcher getMatcher() {
    return this.matcher;
  }

  /**
   * Provides the dictionary
   * 
   * @return the dictionary used for this trial
   */
  public Dictionary getDictionary() {
    return this.dictionary;
  }

  @Override
  public String toString() {
    StringBuffer stringBuffer = new StringBuffer();
    if (this.insert) {
      stringBuffer.append(String.format("%s%n", "ExperimentID: " + this.experimentID));
    }
    stringBuffer.append(String.format("%s%n", "Insert: " + this.insert));
    stringBuffer.append(String.format("%s%n", "Random Seed: " + this.randomSeed));
    stringBuffer.append(String.format("%s%n", "Dataset: " + this.dataset));
    stringBuffer.append(String.format("%s%n", "Fingerprinter:  " + this.fingerprinter));
    stringBuffer.append(String.format("%s%n", "Matcher: " + this.matcher + " (" + this.matcher.getMinimumScore() + ")"));
    String sampleParameterString = (this.sampleParameter < 1) ? this.sampleParameter * 1000 +"%": this.sampleParameter + " (count)";
    stringBuffer.append(String.format("%s%n", "Sample Parameter: " + sampleParameterString + " ("
        + this.sampleFiles.size() + " files)"));
    String dictionaryParameterString = (this.dictionaryParameter < 1) ? this.dictionaryParameter * 1000 +"%": this.dictionaryParameter + " (count)";
    stringBuffer.append(String.format("%s%n", "Dictionary Parameter: " + dictionaryParameterString));
    stringBuffer.append(String.format("%s%n",
        "IDF Range: " + this.dictionary.getMinIDF() + ", " + this.dictionary.getMaxIDF()));
    stringBuffer.append(String.format("%s%n", "Tokenizers:  " + this.dictionary.getTokenizers()));
    stringBuffer.append(String.format("%s%n", "Mangler: " + this.mangler));
    stringBuffer.append(String.format("%s%n", "Splitter: " + this.splitter));
    if (this.file_token != null) {
      stringBuffer.append(String.format("%s%n", "File: " + this.file_token));
      stringBuffer.append(String.format("%s%n",
          "File Bytes: " + this.file_token.start_byte() + "-" + this.file_token.end_byte()));
    }
    if (this.fingerprint != null) {
      stringBuffer.append(String.format("%s%n", "Fingerprint: " + Base64.encode(this.fingerprint)));
    }
    return stringBuffer.toString();
  }
}
