package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.Dictionary;
import java.util.List;
import java.util.Locale;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Token;

/**
 * Splits contents of a text file into tokens based on whitespace or by line. Used for {@linkplain Dictionary} creation.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 **/
public class FileTokenizer extends Tokenizer {

  private static final long serialVersionUID = 7140999692348553037L;

  /** The string specifying to split tokens based on whitespace ("tokens") or by line ("lines") */
  protected String mode;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public FileTokenizer() {
    this.mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string specifying to split tokens based on whitespace or by line
   */
  public FileTokenizer(String mode) {
    setMode(mode);
  }

  /**
   * Sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the file is split
   * based on whitespace.
   * 
   * @param mode
   *          the string specifying to split tokens based on whitespace or by line
   */
  public void setMode(String mode) {
    if (mode.trim().equalsIgnoreCase("tokens") || mode.trim().equalsIgnoreCase("lines")) {
      this.mode = mode.toLowerCase(Locale.ENGLISH).trim();
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("FileTokenizer: Invalid mode (" + mode + ")"));
    }
  }

  /**
   * Provides the token creation mode
   * 
   * @return the string specifying to split tokens based on whitespace or by line
   */
  public String getMode() {
    return this.mode;
  }

  /**
   * Splits the document into tokens
   */
  @Override
  public List<Token> tokenize(String filename) {
    return readFile(filename);
  }

  /**
   * Splits a text document into tokens
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * 
   * @return the vector of tokens present in the file
   */
  protected List<Token> readFile(String filename) {
    BufferedReader buffer_reader = null;
    try {
      // Checks if file is gzip and creates appropriate buffered reader
      buffer_reader = (filename.endsWith(".gz")) ? new BufferedReader(new InputStreamReader(new GZIPInputStream(
          new FileInputStream(filename)), "UTF-8")) : new BufferedReader(new InputStreamReader(new FileInputStream(
          filename), "UTF-8"));

      // buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));

      StringBuffer token = new StringBuffer("");
      int character;
      int byte_start = 0;
      int byte_end = 0;

      synchronized (this.tokenVectorMap) {
        this.tokenVectorMap.clear();
        List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
        synchronized (tokenVector) {
          while ((character = buffer_reader.read()) != -1) {
            if (this.mode.equalsIgnoreCase("lines")) {
              if ('\n' == character || '\r' == character || '\f' == character) {
                if (token.toString().trim().length() > 0) {
                  tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            } else {
              if (Character.isWhitespace(character)) {
                if (token.toString().trim().length() > 0) {
                  tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            }
          }
          buffer_reader.close();
          if (token.toString().trim().length() > 0) {
            tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
          }
          this.tokenVectorMap.put("null", tokenVector);
          return tokenVector;
        }
      }
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      try {
        if (buffer_reader != null) {
          buffer_reader.close();
        }
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return null;
  }
}
