package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.Dictionary;
import java.util.List;
import java.util.Vector;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Token;

/**
 * Uses Oracle's OutsideIn technology to extract text from a file. Splits contents of the file into tokens based on
 * whitespace or by line. Used for {@linkplain Dictionary} creation.
 * 
 * @author Lindsay Neubauer
 */
public class OutsideInFileTokenizer extends FileTokenizer {

  private static final long serialVersionUID = -7648672098261208008L;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public OutsideInFileTokenizer() {
    this.mode = "tokens";
  }

  /**
   * Calls Oracle's Outside In API to extract text from a file and splits that text into tokens
   */
  @Override
  protected List<Token> readFile(String filename) {
    try {
      File tempFile = File.createTempFile("outsideInOutput", null);
      tempFile.deleteOnExit();
      OutputStreamWriter writer = new OutputStreamWriter(new FileOutputStream(tempFile), "UTF-8");
      InputStream inStream = new ProcessBuilder(
          "lib/outsideInExtract", filename).start().getInputStream();
      BufferedReader buffer = new BufferedReader(new InputStreamReader(inStream, "UTF-8"));
      String line;
      while ((line = buffer.readLine()) != null) {
        writer.write(line);
      }
      writer.close();
      inStream.close();
      buffer.close();

      synchronized (this.tokenVectorMap) {
        this.tokenVectorMap.clear();
        Vector<Token> tokenVector = new Vector<Token>();
        synchronized (tokenVector) {
          FileTokenizer fileTokenizer = new FileTokenizer(this.mode);
          fileTokenizer.tokenize(tempFile.getAbsolutePath());
          tokenVector.clear();
          tokenVector.addAll(fileTokenizer.getTokenVectorMap().get("null"));
          this.tokenVectorMap.put("null", tokenVector);
          return tokenVector;
        }
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    return null;
  }
}
