package edu.georgetown.gucs.utility;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.BitSet;
import java.util.Iterator;


/**
 * Converts between bits, bytes, and hex
 * 
 * @author Clay Shields
 * 
 */
public class Converter {

  private static final char[] HEX_CHARS = "0123456789ABCDEF".toCharArray();

  /**
   * Provides a bitset containing the values in a byte array; byte-ordering of the byte array must be big-endian (most
   * significant bit in element 0)
   * 
   * @param bytes
   *          the array of bytes
   * 
   * @return the converted bitset
   * 
   * @see <a href="http://exampledepot.com/egs/java.util/Bits2Array.html">Converting Between a BitSet and a Byte
   *      Array</a>
   */
  public static BitSet byteArrayToBitSet(byte[] bytes) {
    BitSet bits = new BitSet();
    for (int i = 0; i < bytes.length * 8; i++) {
      if ((bytes[bytes.length - i / 8 - 1] & (1 << (i % 8))) > 0) {
        bits.set(i);
      }
    }
    return bits;
  }

  /**
   * Provides a byte array of at least length 1, with the most significant bit guaranteed not to be a 1 (since BitSet
   * does not support sign extension); byte-ordering of the result is big-endian (most significant bit is in element 0);
   * bit at index 0 of the bit set is assumed to be the least significant bit
   * 
   * @param bits
   *          the bitset
   * 
   * @return the converted array of bytes
   */
  public static byte[] bitSetToByteArray(BitSet bits) {
    byte[] bytes = new byte[bits.length() / 8 + 1];
    for (int i = 0; i < bits.length(); i++) {
      if (bits.get(i)) {
        bytes[bytes.length - i / 8 - 1] |= 1 << (i % 8);
      }
    }
    return bytes;
  }

  /**
   * Converts bytes to a hex string
   * 
   * @param buf
   *          the array of bytes
   * 
   * @return the converted string of hex characters
   * 
   * @see <a href="http://forums.xkcd.com/viewtopic.php?f=11&t=16666&p=553936">Java - Converting byte[] to hex
   *      string</a>
   */
  public static String asHex(byte[] buf) {
    char[] chars = new char[2 * buf.length];
    for (int i = 0; i < buf.length; ++i) {
      chars[2 * i] = HEX_CHARS[(buf[i] & 0xF0) >>> 4];
      chars[2 * i + 1] = HEX_CHARS[buf[i] & 0x0F];
    }
    return new String(chars);
  }

  /**
   * Converts a byte array to a set of indices
   * 
   * @param bytes
   *          the byte array
   * 
   * @return the integer array containing the indices
   */
  public static int[] byteArrayToIndices(byte[] bytes) {
    int n = bytes.length * 8;
    int[] tempIndices = new int[n];
    int vals = 0;
    // optimize by skipping zero bytes?
    for (int i = 0; i < n; i++) {
      if ((bytes[bytes.length - i / 8 - 1] & (1 << (i % 8))) > 0) {
        tempIndices[vals] = i;
        vals++;
      }
    }
    int[] indices = new int[vals];
    System.arraycopy(tempIndices, 0, indices, 0, vals);
    return indices;
  }

  /**
   * Computes a SHA-1 hash value indicating the presence or absence of each token in this dictionary
   * 
   * @param tokens
   *          the string iterator over the tokens in this dictionary
   * 
   * @return the hexadecimal string representation of the SHA1 hash of the set of tokens
   * 
   * @see <a href="http://www.anyexample.com/programming/java/java_simple_class_to_compute_sha_1_hash.xml">Java simple
   *      class to compute SHA-1 hash</a>
   */
  public static String SHA1(Iterator<Token> tokens) {
    try {
      MessageDigest md = MessageDigest.getInstance("SHA-1");

      while (tokens.hasNext()) {
        String current = tokens.next().name();
        md.update(current.getBytes("UTF-8"), 0, current.length());
      }
      return Converter.asHex(md.digest());

    } catch (NoSuchAlgorithmException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    return null;
  }
}
