package edu.georgetown.gucs.utility;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;

/**
 * Prints exception messages and stack traces before calling <code>System.exit(-1)</code>
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class ExceptionHandler {

  /**
   * Prints the message and stack trace for this Exception and exits; used with built in java exceptions (not
   * SDText-specific)
   * 
   * @param exception
   *          the exception that provides information on error
   */
  public static void handleExceptionExit(Exception exception) {
    Global.monitor.printErrorMessage("ExceptionHandler", exception.getMessage());
    Global.monitor.printErrorMessage("ExceptionHandler", Arrays.toString(exception.getStackTrace()));
    Global.monitor.writeToBuffer("ExceptionHandler: " + exception.getMessage());
    Global.monitor.writeToBuffer(Arrays.toString(exception.getStackTrace()));
    Global.monitor.closeBuffer();
    System.exit(-1);
  }

  /**
   * Prints the message and stack trace for this Exception and throws a runtime exception; used with exceptions that are
   * SDText-specific (to give user feedback about error)
   * 
   * @param exception
   *          the exception that provides information on error
   */
  public static void handleExceptionThrow(Exception exception) {
    throw new RuntimeException(exception.getMessage());
  }

  /**
   * Prints the message, SQLState, vendor-specific exception code, and stack trace for this SQLException
   * 
   * @param exception
   *          the sqlException that provides information on the database access error or other errors
   */
  public static void handleSQLException(SQLException exception) {
    SQLException sql_exception = exception;
    do {
      String exceptionMessage = String.format("%n%s %n%s %n%s %n%s%n", "SQL Exception:", sql_exception.getMessage(),
          "ANSI-92 SQL State: " + sql_exception.getSQLState(), "Vendor Error Code: " + sql_exception.getErrorCode());
      Global.monitor.printErrorMessage("ExceptionHandler", exceptionMessage);
      Global.monitor.printErrorMessage("ExceptionHandler", Arrays.toString(sql_exception.getStackTrace()));
      Global.monitor.writeToBuffer("ExceptionHandler: " + exceptionMessage);
      Global.monitor.writeToBuffer(Arrays.toString(sql_exception.getStackTrace()));
      sql_exception = sql_exception.getNextException();
    } while (sql_exception != null);
    Global.monitor.closeBuffer();
    System.exit(-1);

  }

  /**
   * Prints the message, SQLState, vendor-specific exception code, and stack trace for this SQLException
   * 
   * @param exception
   *          the sqlException that provides information on the database access error or other errors
   * @param database
   *          the connection for the database that caused this exception
   * @param statement
   *          the SQL statement that caused this exception
   */
  public static void handleSQLException(SQLException exception, Connection database, Statement statement) {
    SQLException sql_exception = exception;
    do {
      String stmt = (statement != null) ? String.format("%nStatement: %s", statement.toString()) : "";
      String exceptionMessage = String.format("%n%s %n%s%s %n%s %n%s%n", "SQL Exception:", sql_exception.getMessage(),
          stmt, "ANSI-92 SQL State: " + sql_exception.getSQLState(), "Vendor Error Code: " + sql_exception.getErrorCode());
      Global.monitor.printErrorMessage("ExceptionHandler", exceptionMessage);
      Global.monitor.printErrorMessage("ExceptionHandler", Arrays.toString(sql_exception.getStackTrace()));
      Global.monitor.writeToBuffer("ExceptionHandler: " + exceptionMessage);
      Global.monitor.writeToBuffer(Arrays.toString(sql_exception.getStackTrace()));
      sql_exception = sql_exception.getNextException();
    } while (sql_exception != null);

    try {
      if (statement != null) {
        statement.close();
      }
      if (database != null) {
        database.close();
      }
    } catch (SQLException ex) {
      Global.monitor.printErrorMessage("ExceptionHandler",
          "Error closing statement or database in ExceptionHandler.handleSQLException");
      handleSQLException(ex);
    }
    Global.monitor.closeBuffer();
    System.exit(-1);
  }
}
