package edu.georgetown.gucs.utility;

import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Creates a list of files in a directory and allows iteration over all and parts of this list
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class FileLister implements Iterable<File> {

  private File directory;
  private ExecutorService executor;
  private BoundedExecutor boundExecutor;
  private LinkedBlockingQueue<File> fileQueue = new LinkedBlockingQueue<File>();
  private Set<File> allFilesSet = Collections.synchronizedSet(new HashSet<File>());
  private List<File> fileList = new Vector<File>();
  private List<File> dictionaryList = new Vector<File>();
  private List<File> sampleList = new Vector<File>();
  private boolean loaded = false;

  /**
   * Constructor that recursively loads files from the given directory and sets the maximum thread count to the number
   * of processors available to the Java virtual machine
   * 
   * @param directory
   *          the string path to the directory to load
   */
  public FileLister(String directory) {
    this.directory = new File(directory);
    int maxThread = Runtime.getRuntime().availableProcessors();
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + maxThread);
    this.executor = Executors.newFixedThreadPool(maxThread);
    this.boundExecutor = new BoundedExecutor(this.executor, maxThread);
    recursiveLoadDirectory(directory);
    shutdownExecutors();
  }

  /**
   * Constructor that recursively loads files from the given directory and sets the maximum thread count
   * 
   * @param directory
   *          the string path to the directory to load
   * @param maxThread
   *          the <code>int</code> maximum threads to use to run this fileLister
   */
  public FileLister(String directory, int maxThread) {
    this.directory = new File(directory);
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + maxThread);
    this.executor = Executors.newFixedThreadPool(maxThread);
    this.boundExecutor = new BoundedExecutor(this.executor, maxThread);
    recursiveLoadDirectory(directory);
    shutdownExecutors();
  }

  /**
   * Constructor that creates a deep copy of the given fileLister object
   * 
   * @param fileLister
   *          the fileLister object to copy
   * @param dictionarySize
   *          the <code>double</code> number of files to include in the dictionary list
   * @param sampleSize
   *          the <code>double</code> number of files to include in the sample list
   */
  public FileLister(FileLister fileLister, double dictionarySize, double sampleSize) {
    Global.monitor.writeToBuffer("FileLister",
        "FileLister(FileLister fileLister, double dictionarySize, double sampleSize)",
        "directory=" + fileLister.getDirectory() + "  dictionarySize=" + dictionarySize + "  sampleSize=" + sampleSize);
    if (!fileLister.getLoaded()) {
      ExceptionHandler.handleExceptionThrow(new Exception(
          "FileLister: Trying to load a FileLister that has not finished loading"));
    }
    this.loaded = fileLister.getLoaded();
    this.directory = fileLister.getDirectory();
    this.allFilesSet = fileLister.getAllFilesSet();
    this.fileQueue = fileLister.getFileQueue();
    this.fileList = fileLister.getFileList();
    this.dictionaryList = fileLister.getDictionaryList();
    this.sampleList = fileLister.getSampleFiles();
    createSublists(dictionarySize, sampleSize);
  }

  /**
   * Provides an iterator over the list of all files
   * 
   * @return the file iterator over the list of all files
   */
  @Override
  public Iterator<File> iterator() {
    return this.fileList.iterator();
  }

  /**
   * Provides an iterator over the list of files used for a sample
   * 
   * @return the file iterator over the list of files used for a sample
   */
  public Iterator<File> sample_iterator() {
    return this.sampleList.iterator();
  }

  /**
   * Provides an iterator over the list of files used for a dictionary
   * 
   * @return the file iterator over the list of files used for a dictionary
   */
  public Iterator<File> dictionary_iterator() {
    return this.dictionaryList.iterator();
  }

  /**
   * Provides the size of the list of all files
   * 
   * @return the <code>int</code> number of files
   */
  public int size() {
    return this.fileList.size();
  }

  /**
   * Provides the size of the list of dictionary files
   * 
   * @return the <code>int</code> number of files used for the dictionary
   */
  public int dictionary_size() {
    return this.dictionaryList.size();
  }

  /**
   * Provides the list of files used for the dictionary
   * 
   * @return the files in the dictionary list
   */
  public List<File> getDictionaryList() {
    return this.dictionaryList;
  }

  /**
   * Provides the list of files used for the sample
   * 
   * @return the files in the sample list
   */
  public List<File> getSampleFiles() {
    return this.sampleList;
  }

  /**
   * Provides the boolean indicating if all files have been loaded into the linked blocking queue
   * 
   * @return the boolean value indicating if all files have been loaded
   */
  public boolean getLoaded() {
    return this.loaded;
  }

  /**
   * Provides the set of all files in the directory
   * 
   * @return the set of all the files in the directory
   */
  public Set<File> getAllFilesSet() {
    return this.allFilesSet;
  }

  /**
   * Provides the linked blocking queue of all the files in the directory
   * 
   * @return the linked blocking queue of all the files in the directory
   */
  public LinkedBlockingQueue<File> getFileQueue() {
    return this.fileQueue;
  }

  /**
   * Provides a list of all files in the directory
   * 
   * @return the list of all the files in the directory
   */
  public List<File> getFileList() {
    return this.fileList;
  }

  /**
   * Determines whether all the files have been loaded and unloaded from the queue of files
   * 
   * @return the <code>boolean</code> value whether all the files have been loaded and unloaded from the queue of files
   */
  public synchronized boolean loadComplete() {
    return (this.loaded && this.fileQueue.size() == 0);
  }

  /**
   * Provides the path to the directory used for the initial list of files
   * 
   * @return the string path to the directory used to initially create this list of files
   */
  public File getDirectory() {
    return this.directory;
  }

  /**
   * Retrieves and removes the head of the file queue, waiting if necessary until an element becomes available
   * 
   * @return the file at the head of the file queue
   */
  public synchronized File takeFile() {
    try {
      return this.fileQueue.take();
    } catch (InterruptedException e) {
      Global.monitor.writeToBuffer("FileLister", "takeFile", "thread_id = " + Thread.currentThread().getId());
      Global.monitor.writeToBuffer("FileLister", "takeFile", Arrays.toString(Thread.currentThread().getStackTrace()));
      ExceptionHandler.handleExceptionExit(e);
    }
    return null;
  }

  /**
   * Adds the file to the tail of the file queue, waiting if necessary for space to become available
   * 
   * @param file
   *          the file to put at the tail of the file queue
   */
  protected synchronized void putFile(File file) {
    try {
      this.fileQueue.put(file);
    } catch (InterruptedException e) {
      Global.monitor.writeToBuffer("FileLister", "putFile", "thread_id = " + Thread.currentThread().getId());
      Global.monitor.writeToBuffer("FileLister", "putFile", Arrays.toString(Thread.currentThread().getStackTrace()));
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Adds the file to the end of this list of files used for a dictionary
   * 
   * @param file
   *          the file to add to this list of files
   */
  protected void addFile(File file) {
    synchronized (this.allFilesSet) {
      if (this.allFilesSet.add(file)) {
        this.fileList.add(file);
      }
    }
  }

  /**
   * Returns a list of each of the files in this fileLister, each on a separate line
   */
  @Override
  public String toString() {
    StringBuffer string = new StringBuffer();
    for (File file : this.fileList) {
      string.append(String.format("%s%n", file.toString()));
    }
    return string.toString();
  }

  /**
   * Recursively loads all the files from the given directory into the list of files used for a dictionary
   * 
   * @param path
   *          the string path to the directory to load
   */
  private void recursiveLoadDirectory(String path) {
    Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "directory=" + path);

    File root = new File(path);
    for (File file : Arrays.asList(root.listFiles())) {
      if (file.isFile() && file.canRead()) {
        // ignore .DS_Store files
        if (file.toString().indexOf(".DS_Store") < 0) {
          FileListerWorker currentFile = new FileListerWorker(file, this);
          this.boundExecutor.submitTask(currentFile);
        }
      } else if (file.isDirectory() && file.canRead()) {
        // Ignore .svn directories
        if (file.toString().indexOf(".svn") < 0) {
          recursiveLoadDirectory(file.toString());
        }
      }
    }
    Global.monitor.printMessage("FileLister", "Files Loaded: " + root.getAbsolutePath());
    Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "loaded=" + this.loaded);
  }

  private boolean shutdownExecutors() {
    this.executor.shutdown();
    while (!this.executor.isTerminated()) {
      try {
        // Wait a second and see if we are done
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "thread_id = "
            + Thread.currentThread().getId());
        Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory",
            Arrays.toString(Thread.currentThread().getStackTrace()));
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    this.loaded = true;
    Global.monitor.printMessage("FileLister", "Executors shutdown");
    Global.monitor.writeToBuffer("FileLister", "shutdownExecutors", this.executor.toString());
    return true;
  }

  /**
   * Populates the dictionary and sample lists, waiting for all files to be loaded before randomly assigning files to
   * one of the lists. Dictionary and sample lists do not share any files. If this method is never called all files are
   * assigned to the dictionary list.
   * 
   * @param dictionarySize
   *          the <code>double</code> number of files to include in the dictionary list
   * @param sampleSize
   *          the <code>double</code> number of files to include in the sample list
   */
  public void createSublists(double dictionarySize, double sampleSize) {
    if (dictionarySize < 0 || sampleSize < 0) {
      ExceptionHandler.handleExceptionThrow(new Exception("FileLister: Negative size parameter(s): dictionarySize = "
          + dictionarySize + "  sampleSize = " + sampleSize));
    }
    int dictionaryCount = (int) ((dictionarySize >= 1) ? dictionarySize
        : ((dictionarySize * 10 * this.fileList.size()) + .5));
    int sampleCount = (int) ((sampleSize >= 1) ? sampleSize : ((sampleSize * 10 * this.fileList.size()) + .5));
    Global.monitor.printMessage("FileLister", "createSublists: dictionaryCount=" + dictionaryCount + "  sampleCount="
        + sampleCount);
    Global.monitor.writeToBuffer("FileLister", "createSublists", "dictionaryCount=" + dictionaryCount);
    Global.monitor.writeToBuffer("FileLister", "createSublists", "sampleCount=" + sampleCount);

    boolean loop = true;
    while (loop) {
      if (this.loaded) {
        loop = false;
        if (dictionaryCount > this.fileList.size() - sampleCount
            || sampleCount > this.fileList.size() - dictionaryCount) {
          ExceptionHandler.handleExceptionThrow(new Exception(String.format("%s%n  %s%n  %s%n  %s%n",
              "FileLister: Incompatible parameters: dictionary files + sample files <= total directory files",
              "Total directory files: " + this.fileList.size(), "Requested number of dictionary files: "
                  + dictionaryCount, "Requested number of sample files: " + sampleCount)));
        }

        Collections.shuffle(this.fileList);
        this.dictionaryList.clear();
        this.sampleList.clear();
        this.dictionaryList.addAll(this.fileList.subList(0, dictionaryCount));
        this.sampleList.addAll(this.fileList.subList(dictionaryCount, dictionaryCount + sampleCount));
        Global.monitor.printMessage("FileLister", "createSublists: dictionaryList.size=" + this.dictionaryList.size()
            + "  sampleList.size=" + this.sampleList.size());
      }
    }
    Global.monitor.printMessage("FileLister", "createSublists: completed");

  }
}
